@@ -24,12 +24,17 @@ filter.set_name("Images")
 filter.add_mime_type("image/png")
 filter.add_mime_type("image/jpeg")
 filter.add_mime_type("image/gif")
+filter.add_pattern("*.png")
+filter.add_pattern("*.jpg")
+filter.add_pattern("*.gif")
+filter.add_pattern("*.tif")
+filter.add_pattern("*.xpm")
 dialog.add_filter(filter)
 
 response = dialog.run()
-dialog.destroy()
 if response == gtk.RESPONSE_OK:
     print dialog.get_filename(), 'selected'
 elif response == gtk.RESPONSE:
     print 'Closed, no files selected'
+dialog.destroy()
     
