@@ -66,7 +66,9 @@ _pygtk_style_helper_new(GtkStyle *style, int type, gpointer array)
 						   &PyGtkStyleHelper_Type);
     if (self == NULL)
 	return NULL;
+    pyg_unblock_threads();
     self->style = g_object_ref(style);
+    pyg_block_threads();
     self->type = type;
     self->array = array;
     return (PyObject *)self;
@@ -148,12 +150,12 @@ pygtk_style_helper_setitem(PyGtkStyleHelper_Object *self, int pos,
 		PyErr_SetString(PyExc_TypeError, "can only assign a GdkGC");
 		return -1;
 	    }
+	    pyg_unblock_threads();
 	    if (array[pos]) {
-		pyg_unblock_threads();
 		g_object_unref(array[pos]);
-		pyg_block_threads();
 	    }
 	    array[pos] = GDK_GC(g_object_ref(pygobject_get(value)));
+	    pyg_block_threads();
 	    return 0;
 	}
     case STYLE_PIXMAP_ARRAY:
@@ -165,15 +167,15 @@ pygtk_style_helper_setitem(PyGtkStyleHelper_Object *self, int pos,
 				"can only assign a GdkPixmap or None");
 		return -1;
 	    }
+	    pyg_unblock_threads();
 	    if (array[pos]) {
-		pyg_unblock_threads();
 		g_object_unref(array[pos]);
-		pyg_block_threads();
 	    }
 	    if (value != Py_None)
 		array[pos] = GDK_PIXMAP(g_object_ref(pygobject_get(value)));
 	    else
 		array[pos] = NULL;
+	    pyg_block_threads();
 	    return 0;
 	}
     }
@@ -1092,18 +1094,70 @@ static PyTypeObject PyGtkTreeModelRowIter_Type = {
     (iternextfunc)pygtk_tree_model_row_iter_next
 };
 
+int
+_pygtk_tree_model_remove_row(GtkTreeModel *model, GtkTreeIter *iter)
+{
+    GtkTreeModel *child;
+    GtkTreeIter citer;
+
+    if (GTK_IS_LIST_STORE(model)) {
+        gtk_list_store_remove(GTK_LIST_STORE(model), iter);
+        return 0;
+    }
+
+    if (GTK_IS_TREE_STORE(model)) {
+        gtk_tree_store_remove(GTK_TREE_STORE(model), iter);
+        return 0;
+    }
+
+    if (GTK_IS_TREE_MODEL_SORT(model)) {
+        child = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(model));
+        gtk_tree_model_sort_convert_iter_to_child_iter(
+            GTK_TREE_MODEL_SORT(model), &citer, iter);
+        return _pygtk_tree_model_remove_row(child, &citer);
+    }
+
+    if (GTK_IS_TREE_MODEL_FILTER(model)) {
+        child = gtk_tree_model_filter_get_model(GTK_TREE_MODEL_FILTER(model));
+        gtk_tree_model_filter_convert_iter_to_child_iter(
+            GTK_TREE_MODEL_FILTER(model), &citer, iter);
+        return _pygtk_tree_model_remove_row(child, &citer);
+    }
+
+    PyErr_SetString(PyExc_TypeError,
+                    "cannot remove rows in this tree model");
+    return -1;
+}
+
 int
 _pygtk_tree_model_set_row(GtkTreeModel *model, GtkTreeIter *iter,
 			  PyObject *items)
 {
     gint n_columns, i;
+    GtkTreeModel *child;
+    GtkTreeIter citer;
 
-    if (!GTK_IS_LIST_STORE(model) && !GTK_IS_TREE_STORE(model)) {
+    if (!GTK_IS_LIST_STORE(model) && !GTK_IS_TREE_STORE(model) &&
+        !GTK_IS_TREE_MODEL_SORT(model) && !GTK_IS_TREE_MODEL_FILTER(model)) {
 	PyErr_SetString(PyExc_TypeError,
-			"can not set cells in this tree model");
+			"cannot set cells in this tree model");
 	return -1;
     }
 
+    if (GTK_IS_TREE_MODEL_SORT(model)) {
+        child = gtk_tree_model_sort_get_model(GTK_TREE_MODEL_SORT(model));
+        gtk_tree_model_sort_convert_iter_to_child_iter(
+            GTK_TREE_MODEL_SORT(model), &citer, iter);
+        return _pygtk_tree_model_set_row(child, &citer, items);
+    }
+
+    if (GTK_IS_TREE_MODEL_FILTER(model)) {
+        child = gtk_tree_model_filter_get_model(GTK_TREE_MODEL_FILTER(model));
+        gtk_tree_model_filter_convert_iter_to_child_iter(
+            GTK_TREE_MODEL_FILTER(model), &citer, iter);
+        return _pygtk_tree_model_set_row(child, &citer, items);
+    }
+
     if (!PySequence_Check(items)) {
 	PyErr_SetString(PyExc_TypeError, "expecting a sequence");
 	return -1;
